#! /usr/bin/env python

from __future__ import print_function

import nids
import sys
from pwny import *


target.arch = Target.Arch.unknown
target.endian = Target.Endian.big


def parse_frag(data):
    # Parse fragment returning its data, whether its the last fragment and
    # the rest of the data.
    hdr = U32(data[:4])
    frag_len = hdr & (2**31-1)
    frag_last = hdr & 1<<31
    frag_data = data[4:4 + frag_len]
    data = data[4 + frag_len:]
    return frag_data, frag_last, data


def parse_int(data):
    # Parse a 32bit integer. Returns integer and rest of data.
    return U32(data[:4]), data[4:]


def parse_long(data):
    # Parse a 64bit integer. Returns integer and rest of data.
    return U64(data[:8]), data[8:]


def parse_guid(data):
    # Parse a guid. Return guid as bytes and rest of data.
    return data[:16], data[16:]


def parse_str(data):
    # Parse a string. Return string as bytes and rest of data.
    str_len, data = parse_int(data)
    pad = (4 - (str_len % 4)) % 4
    return data[:str_len], data[str_len + pad:]


def parse_dict(data):
    # Parse a dictionary. Returns dict and rest of data.
    dict_size, data = parse_int(data)
    if not dict_size:
        return {}, data
    items, data = parse_int(data)
    result = {}
    for i in range(items):
        key_len, data = parse_int(data)
        value_len, data = parse_int(data)
        key, data = data[:key_len], data[key_len + 1:]
        value, data = data[:value_len], data[value_len:]
        result[key] = value
    return result, data


def parse_creds(data):
    # Parse RPC credentials. Return flavor, content and rest of data.
    flavor, data = parse_int(data)
    length, data = parse_int(data)
    return flavor, data[:length], data[length:]


def parse_verifier(data):
    # Parse RPC verifier block. Return flavor, content and rest of data.
    flavor, data = parse_int(data)
    length, data = parse_int(data)
    return flavor, data[:length], data[length:]


# Total set of data to write. Dictionary key should be filename,
# content a list of (stripe, offset, data) tuples.
files = {}


def handle_gluster_stream(data):
    # Parse the collected data from a single stream of glusterfs client-server communication.

    # The packet data.
    pkt = b''

    # Files opened in this stream. Maps GFID to (filename, stripe).
    local_files = {}

    while data:
        frag_data, frag_last, data = parse_frag(data)
        pkt += frag_data
        if not frag_last:
            continue

        # Parse the RPC header
        xid, pkt = parse_int(pkt)
        mtype, pkt = parse_int(pkt)
        rpc_ver, pkt = parse_int(pkt)
        program, pkt = parse_int(pkt)
        program_ver, pkt = parse_int(pkt)
        procedure, pkt = parse_int(pkt)
        creds_flavor, creds, pkt = parse_creds(pkt)
        verifier_flavor, verifier, pkt = parse_verifier(pkt)

        if program_ver == 330 and procedure == 23:
            # Handle file creation. Register filename + stripe.
            guid, pkt = parse_guid(pkt)
            flags, pkt = parse_int(pkt)
            mode, pkt = parse_int(pkt)
            umask, pkt = parse_int(pkt)
            fn, pkt = parse_str(pkt)
            d, pkt = parse_dict(pkt)
            stripe_index = int(d['trusted.gv0-stripe-0.stripe-index'].rstrip('\x00'))
            gfid = d['gfid-req']
            local_files[gfid] = (fn, stripe_index)
            files.setdefault(fn, [])
        elif program_ver == 330 and procedure == 13:
            # Write data. Use GFID to determine filename + stripe.
            gfid, pkt = parse_guid(pkt)
            fd, pkt = parse_long(pkt)
            offset, pkt = parse_long(pkt)
            chunk_size, pkt = parse_int(pkt)
            flags, pkt = parse_int(pkt)
            d, pkt = parse_dict(pkt)
            fn, stripe_index = local_files[gfid]
            files[fn].append((stripe_index, offset, pkt))

        pkt = b''


end_states = (nids.NIDS_CLOSE, nids.NIDS_TIMEOUT, nids.NIDS_RESET)
def handle_tcp_stream(tcp):
    if tcp.nids_state == nids.NIDS_JUST_EST:
        # We only care about the 'server' stream/
        tcp.server.collect = 1
    elif tcp.nids_state == nids.NIDS_DATA:
    	tcp.discard(0)
    elif tcp.nids_state in end_states:
        # We pass each individual stream to handle_gluster_stream.
        handle_gluster_stream(tcp.server.data[:tcp.server.count])


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print('Usage: %s <pcap>' % sys.argv[0], file=sys.stderr)
        sys.exit(1)

    # Set up nids.
    nids.param("scan_num_hosts", 0)         # disable portscan detection
    nids.chksum_ctl([('0.0.0.0/0', False)]) # disable checksumming
    nids.param('filename', sys.argv[1])
    nids.init()

    # Run stream through the handler.
    print('Parsing stream.. Please wait..')
    nids.register_tcp(handle_tcp_stream)
    nids.run()

    # Write chunks to disk. Sort by (offset, stripe).
    for filename, chunks in files.items():
        print('Writing', filename)
        with open(filename, 'wb') as f:
            for stripe, offset, data in sorted(chunks, key=lambda c: (c[1], c[0])):
                f.write(data)
